@@ -591,6 +591,13 @@ namespace Google.Protobuf
             Assert.Throws<InvalidOperationException>(() => JsonFormatter.Default.Format(any));
         }
 
+        [Test]
+        public void AnyEmpty()
+        {
+            var any = new Any();
+            AssertJson("{ }", JsonFormatter.Default.Format(any));
+        }
+
         [Test]
         [TestCase(typeof(BoolValue), true, "true")]
         [TestCase(typeof(Int32Value), 32, "32")]
