@@ -12,16 +12,12 @@
 
 #include "absl/container/flat_hash_set.h"
 #include "absl/log/absl_check.h"
-#include "absl/log/absl_log.h"
 #include "absl/memory/memory.h"
 #include "absl/strings/cord.h"
 #include "absl/strings/string_view.h"
 #include "absl/strings/substitute.h"
 #include "google/protobuf/compiler/code_generator.h"
-#include "google/protobuf/compiler/code_generator_lite.h"
 #include "google/protobuf/compiler/plugin.h"
-#include "upb/base/status.hpp"
-#include "upb/base/string_view.h"
 #include "upb/mem/arena.hpp"
 #include "upb/reflection/def.hpp"
 #include "upb_generator/common.h"
